#!/usr/bin/env python3

import os
import sys
import json
import joblib
import hashlib
import logging
import tempfile
import numpy as np
import pandas as pd
from . import utils, simulate, test, train


logger = logging.getLogger(__name__)


def train_cli(config: str):
    """ Train a model """
    config = utils.Config(config).config
    # Set global seed
    np.random.seed(config['seed'])
    # Read data and map DNA target label to 1
    data = pd.read_csv(config['input'], encoding=config['encoding'])
    assert len(data[config['target']].unique()) == 2
    assert config['DNAclass'] in data[config['target']].unique()
    data[config['target']] = data[config['target']].apply(
        lambda x: 1 if x == config['DNAclass'] else 0)
    data = train.splitData(data, **config)
    # Train model
    models = train.trainModel(data, **config)
    # Write trained logistic and catboost models and parameter definitions
    for name, model in models.items():
        joblib.dump(model['model'], f'{config["out"]}/{name}-trained.pkl')
        with open(f'{config["out"]}/{name}-params.json', 'w') as fh:
            json.dump(model['params'], fh)
    # Write split data
    for name, df in data.items():
        df.to_pickle(f'{config["out"]}/{name}.pkl')


def test_cli(config: str):
    """ Test a pre-trained model. """
    config = utils.Config(config).config
    np.random.seed(config['seed']) # Set global seed
    data = _readData(config)
    models = {'catboost': {}, 'logistic': {}}
    for name in models:
        models[name]['model'] = joblib.load(
            f'{config["out"]}/{name}-trained.pkl')
        with open(f'{config["out"]}/{name}-params.json') as fh:
            models[name]['params'] = json.load(fh)

    featureImportances = test.getFeatureImportance(models['catboost']['model'])
    fig = featureImportances.plot.barh()
    fig.figure.savefig(f'{config["out"]}/featureImportances.png', dpi=300)

    fig, ax = test.plotROC(models, data)
    fig.figure.savefig(f'{config["out"]}/ROCcurve.png', dpi=300)

    fig, ax = test.plotPrecisionRecall(models, data)
    fig.figure.savefig('PRcurve.png', dpi=300)

    fig, ax = test.plotCalibrationCurve(models, data, strategy='quantile')
    fig.figure.savefig('CalibrationCurve.png', dpi=300)

    for name in models:
        report = test.evaluate(models[name]['model'], data)
        with open(f'{config["out"]}/{name}-report.json', 'w') as fh:
            json.dump(report, fh)


def retrain_cli(config: str):
    """ Re-train model on full dataset """
    config = utils.Config(config).config
    np.random.seed(config['seed']) # Set global seed
    data = _readData(config)
    modelType = config['finalModel']
    if modelType not in ['catboost', 'logistic']:
        if modelType:
            logging.error(
                f'Invalid configuration for "finalModel" - {finalModel}')
        logging.error(
            'Please set "finalModel" to either "catboost" or "logistic"')
    model = joblib.load(
        f'{config["out"]}/{modelType}-trained.pkl')
    model = train.refitData(model, data)
    joblib.dump(model, f'{config["out"]}/{modelType}-final.pkl')


def predict_cli(
        data, model, verify: bool = False,
        encoding: str = 'utf-8', sep: str = ',', out=sys.stdout):
    data = pd.read_csv(data, encoding=encoding, sep=sep)
    model = joblib.load(model)
    data[['Attend_prob', 'DNA_prob', 'Prediction']] = (
        test.predict(model, data))
    data['Prediction'] = data['Prediction'].map({1: 'DNA', 0: 'Attend'})
    if out is not None:
        data.to_csv(out)
    if verify:
        if verifyHash(data):
            print('Success: output matches expected hash.', file=sys.stderr)
            return 0
        else:
            print('Error: output does NOT match expected hash', file=sys.stderr)
            return 1


def verifyHash(df: str, readSize: int = 4096):
    tf = tempfile.NamedTemporaryFile(delete=False)
    df.to_csv(tf.name, float_format='%.3f')
    sha256Hash = hashlib.sha256()
    with open(tf.name, 'rb') as f:
        data = f.read(readSize)
        while data:
            sha256Hash.update(data)
            data = f.read(readSize)
    hash = sha256Hash.hexdigest()
    os.remove(tf.name)
    return hash == ('801cb98641256aaf8c3a57c7af87da80'
                    'bd2f92088a8c9dcd4609c06e03395484')


def simulate_cli(
        config: str, size: int = 50_000,
        noise: float = 0.2, seed: int = 42, out=sys.stdout):
    """ Randomly generate some example data """
    assert seed > 0
    assert size > 0
    assert noise >= 0
    # Randomly generate some artificial attendance data
    df = simulate.generateData(size, seed, noise)
    df.to_csv(out, index=False)
    simulate.writeConfig(config)


def _readData(config):
    """ Read pre-split data """
    data = {}
    for name in ['X_train', 'y_train', 'X_test', 'y_test', 'X_val', 'y_val']:
        data[name] = pd.read_pickle(f'{config["out"]}/{name}.pkl')
    return data



#!/usr/bin/env python3

_MAJOR_VERSION = '0'
_MINOR_VERSION = '1'
_PATCH_VERSION = '1'

_VERSION_SUFFIX = None

# Example, '0.1.0-dev'
__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])

if _VERSION_SUFFIX:
    __version__ = f'{__version__}-{_VERSION_SUFFIX}'



#!/usr/bin/env python3

import dnattend.utils as utils
import logging
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from scipy.stats import randint, uniform
from category_encoders.woe import WOEEncoder
from catboost import CatBoostClassifier, Pool
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import balanced_accuracy_score
from sklearn.calibration import CalibratedClassifierCV
from sklearn.model_selection import RandomizedSearchCV
from sklearn.utils.class_weight import compute_class_weight
from sklearn.model_selection import train_test_split, KFold
from sklearn.metrics import roc_curve, precision_recall_curve

logger = logging.getLogger(__name__)


def splitData(
        data, target, train_size: float = 0.8, test_size: float = 0.1,
        val_size: float = 0.1, **kwargs):
    """ Split data into test / train / validation """

    assert train_size + test_size + val_size == 1
    logger.info(f'Splitting data: train ({train_size:.1%}) : '
                f'test ({test_size:.1%}) : validation ({val_size:.1%}).')

    X = data.copy()
    y = X.pop(target)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=(1 - train_size))

    split2Size = test_size / (test_size + val_size)
    X_val, X_test, y_val, y_test = train_test_split(
        X_test, y_test, test_size=split2Size)

    return ({
        'X_train': X_train.copy(), 'y_train': y_train.copy(),
        'X_test': X_test.copy(), 'y_test': y_test.copy(),
        'X_val': X_val.copy(), 'y_val': y_val.copy()
    })


def _getClassWeights(y):
    # Set class weight to balanced probalities for more easy interpretation
    classes = np.unique(y)
    weights = compute_class_weight(
        class_weight='balanced', classes=classes, y=y)
    return  dict(zip(classes, weights))


def _buildPreProcessor(catCols, numericCols, boolCols, mode: str = 'catboost'):
    catPipe = SimpleImputer(strategy='constant')
    if mode == 'logistic':
        catPipe =  Pipeline([('impute', catPipe), ('encode', WOEEncoder())])
    transformers = ([
        ('categories', catPipe, catCols),
        ('numeric',    SimpleImputer(strategy='mean'), numericCols),
        ('boolean',    SimpleImputer(strategy='most_frequent'), boolCols),
    ])
    featureTransformer = ColumnTransformer(
        transformers=transformers, remainder='drop')
    if mode == 'logistic':
        preProcessor = Pipeline(steps=[
            ('prepare',         utils._prepareData(catCols, numericCols, boolCols)),
            ('columnTransform', featureTransformer),
            ('scale',           StandardScaler()),
    ])
    else:
        preProcessor = Pipeline(steps=[
            ('prepare',         utils._prepareData(catCols, numericCols, boolCols)),
            ('columnTransform', featureTransformer),
        ])
    return preProcessor


def trainModel(
        data: dict,
        catCols: list = None,
        numericCols: list = None,
        boolCols: list = None,
        cvFolds: int = 5,
        catboostIterations: int = 100,
        hypertuneIterations: int = 5,
        evalIterations: int = 10000,
        earlyStoppingRounds: int = 10,
        hyperParams: dict = None,
        tuneThresholdBy: str = 'f1',
        verbose: int = 0,
        nJobs: int = 1,  **kwargs):

    for colList in [catCols, numericCols, boolCols]:
        colList = [] if colList is None else colList
        for col in colList.copy():
            if col not in data['X_train']:
                logger.error(f'Feature "{col}" not in data - removing.')
                colList.remove(col)
    if (not catCols) and (not numericCols) and (not boolCols):
        raise ValueError('No features provided.')
    assert tuneThresholdBy in ['f1', 'roc']

    models = {}
    for mode in ['logistic', 'catboost']:
        if mode == 'catboost':
            model, params = _trainCatBoost(
                data, catCols, numericCols,
                boolCols, cvFolds, catboostIterations,
                hypertuneIterations, evalIterations,
                earlyStoppingRounds, hyperParams,
                verbose, nJobs)
        else:
            model, params = _trainLogistic(data, catCols, numericCols, boolCols)
        params = _fixParams(params)
        model = _rebuildPipeline(model, mode)
        _ = model.set_params(**params)
        model = refitData(model, data, noTest=True)

        # Tune decision threshold to balance FPR and TPR
        logger.info('Optimising decision threshold.')
        threshold = _tuneThreshold(
            model, data['X_train'], data['y_train'], tuneThresholdBy)

        logger.info(f'Setting decision threshold to {threshold:.3f}.')
        params['preprocess__prepare__decisionThreshold'] = threshold
        _ = model.set_params(
            preprocess__prepare__decisionThreshold=threshold,
        )
        models[mode] = {'model': model, 'params': params}
    return models


def _trainLogistic(
        data: dict, catCols: list = None,
        numericCols: list = None, boolCols: list = None):

    logger.info('Running base logistic regression model...')
    preProcessor = _buildPreProcessor(
        catCols, numericCols, boolCols, mode='logistic')
    baseEstimator = LogisticRegression(class_weight='balanced')
    model = Pipeline(steps=[
        ('preprocess',     preProcessor),
        ('estimator',      baseEstimator),
    ])
    model = model.fit(data['X_train'], data['y_train'])
    params = {}
    return model, params


def _trainCatBoost(
        data: dict,
        catCols: list = None,
        numericCols: list = None,
        boolCols: list = None,
        cvFolds: int = 5,
        catboostIterations: int = 100,
        hypertuneIterations: int = 5,
        evalIterations: int = 10000,
        earlyStoppingRounds: int = 10,
        hyperParams: dict = None,
        verbose: int = 0,
        nJobs: int = 1):

    if (hyperParams is None) or (not isinstance(hyperParams, dict)):
        hyperParams = ({
            'estimator__depth':           randint(4, 10),
            'estimator__l2_leaf_reg':     randint(2, 10),
            'estimator__random_strength': uniform.rvs(0, 10, size=100),
        })

    logger.info('Building catboost classifier...')
    logger.info('Constructing pre-processing + estimator Pipeline.')
    preProcessor = _buildPreProcessor(
        catCols, numericCols, boolCols, mode='catboost')

    logger.info('Estimating class weights from data.')
    class_weights = _getClassWeights(data['y_train'])

    # Combine processor and modelling steps into a Pipeline object
    catColIdx = preProcessor.named_steps['prepare'].catColIdx
    baseEstimator = CatBoostClassifier(
        cat_features=catColIdx,
        eval_metric='Logloss',
        class_weights=class_weights,
        allow_writing_files=False,
        iterations=catboostIterations, verbose=verbose)
    model = Pipeline(steps=[
        ('preprocess',     preProcessor),
        ('estimator',      baseEstimator),
    ])
    logger.info(f'Performing {cvFolds}-fold cross-validated random search '
                 f'of hyper-parameters ({hypertuneIterations} iterations).')
    gridSearch = RandomizedSearchCV(
        model, hyperParams, scoring='neg_log_loss', cv=cvFolds,
        refit=False, n_jobs=nJobs, n_iter=hypertuneIterations, verbose=verbose)
    _ = gridSearch.fit(data['X_train'], data['y_train'])

    # Extract best parameters from cross-validated randomised search
    params = gridSearch.best_params_
    params['estimator__iterations'] = evalIterations
    _ = model.set_params(**params)

    # Pre-process the validation set with the tuned model parameters.
    # Required since eval_set is other not processed before CatBoost
    fitPreProcessor = model.named_steps['preprocess'].fit(
        data['X_train'], data['y_train'])
    X_val = fitPreProcessor.transform(data['X_val'])
    evalSet = Pool(X_val, data['y_val'], cat_features=catColIdx)
    logger.info('Re-fitting tuned model to estimate '
                 'optimal iterations using Early Stopping.')

    _ = model.fit(
        data['X_train'], data['y_train'], estimator__eval_set=evalSet,
        estimator__early_stopping_rounds=earlyStoppingRounds)

    # Update iteration parameter to optimal and write to file
    bestIteration = model.named_steps['estimator'].get_best_iteration()
    logger.info(f'Setting iterations to {bestIteration}.')
    params['estimator__iterations'] = bestIteration

    return model, params


def _tuneThreshold(model, X_train, y_train, mode):
    assert mode in ['f1', 'roc']
    y_trainInt = y_train.apply(lambda x: 1 if x == model.classes_[1] else 0)
    trainPredictProb = model.predict_proba(X_train)[:,1]
    if mode == 'f1':
        logger.info('Tuning threshold by F1-score.')
        precision, recall, thresholds = precision_recall_curve(
            y_trainInt, trainPredictProb)
        fscore = (2 * precision * recall) / (precision + recall)
        idx = np.nanargmax(fscore)
        optimalThreshold = thresholds[idx]
    else:
        logger.info('Tuning threshold by ROC.')
        fpr, tpr, thresholds = roc_curve(
            y_trainInt, trainPredictProb, drop_intermediate=False)
        idx = np.nanargmin(np.abs(fpr + tpr - 1))
        optimalThreshold = thresholds[idx]
    return optimalThreshold


def _fixParams(params: dict):
    """ Fix the path to parameters """
    newParams = {}
    for param, val in params.items():
        if param.startswith('estimator__'):
            param = 'estimator__base_estimator' + param[9:]
        newParams[param] = val
    return newParams


def _rebuildPipeline(model, mode: str = 'catboost'):
    """ Rebuild pipeline with  CalibratedClassifierCV """
    catCols = model.get_params()['preprocess__prepare__catCols']
    numericCols = model.get_params()['preprocess__prepare__numericCols']
    boolCols = model.get_params()['preprocess__prepare__boolCols']
    preProcessor = _buildPreProcessor(catCols, numericCols, boolCols, mode)
    catColIdx = preProcessor.named_steps['prepare'].catColIdx
    if mode == 'catboost':
        class_weights = model.get_params()['estimator__class_weights']
        estimator = CatBoostClassifier(
            cat_features=catColIdx, eval_metric='Logloss',
            class_weights=class_weights, verbose=0,
            allow_writing_files=False)
    else:
        estimator = LogisticRegression(
            class_weight='balanced', verbose=0)
    model = Pipeline(steps=[
        ('preprocess',     preProcessor),
        ('estimator',      CalibratedClassifierCV(estimator)),
    ])
    return model


def refitData(model, data, noTest=False):
    """ Perform final refit with full data """
    if noTest:
        logger.info('Refitting model with tuned parameters '
                    'on training + validation dataset.')
        X = [data['X_train'], data['X_val']]
        y = [data['y_train'], data['y_val']]
    else:
        logger.info('Refitting model with tuned parameters on full dataset.')
        X = [data['X_train'], data['X_test'], data['X_val']]
        y = [data['y_train'], data['y_test'], data['y_val']]
    logger.info('Recalibrating probabilities with cross-validation.')
    model.fit(pd.concat(X), pd.concat(y))
    return model



#!/usr/bin/env python3


import logging
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.metrics import classification_report
from sklearn.calibration import CalibrationDisplay
from sklearn.metrics import roc_curve, roc_auc_score, RocCurveDisplay
from sklearn.metrics import precision_recall_curve, PrecisionRecallDisplay


logger = logging.getLogger(__name__)


def getFeatureImportance(model):
    calClassifiers = model.named_steps['estimator'].calibrated_classifiers_
    importances = 0
    for classifier in calClassifiers:
        importances += classifier.base_estimator.feature_importances_
    importances /= len(calClassifiers)
    # model.named_steps['estimator'].base_estimator.feature_importances_
    importances = pd.Series(
        importances,
        model.named_steps['preprocess'].named_steps['prepare'].validCols
    ).sort_values(ascending=True)
    return importances


def _processTestData(model, data):
    y_test = data['y_test'].apply(
        lambda x: 1 if x == model.classes_[1] else 0)
    test_pred_proba = model.predict_proba(data['X_test'])[:,1]
    return y_test, test_pred_proba


def plotROC(models, data, figsize=None):
    fig, ax = plt.subplots(figsize=figsize)
    labels = []
    for name in ['logistic', 'catboost']:
        model = models[name]['model']
        y_test, test_pred_proba = _processTestData(model, data)
        AUC = roc_auc_score(y_test, test_pred_proba)
        fpr, tpr, thresholds = roc_curve(
            y_test, test_pred_proba, drop_intermediate=False)
        idx = np.nanargmin(np.abs(fpr + tpr - 1))
        RocCurveDisplay.from_estimator(
            model, data['X_test'], data['y_test'], ax=ax)
        if name == 'catboost':
            ax.axhline(tpr[idx], xmax=fpr[idx], ls='--', alpha=0.5, c='black')
            ax.axvline(fpr[idx], ymax=tpr[idx], ls='--', alpha=0.5, c='black')
            ax.scatter(fpr[idx], tpr[idx], c='black')
        label = f'{name}: AUC = {AUC:.2f}, Threshold = {thresholds[idx]:.3f}'
        labels.append(label)
    ax.axline((0, 0), slope=1, ls='--', color='red')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.set_xlabel('False Positive Rate')
    ax.set_ylabel('True Positive Rate')
    ax.legend(labels=labels, loc='lower right')
    return fig, ax


def plotPrecisionRecall(models, data, figsize=None):
    fig, ax = plt.subplots(figsize=figsize)
    labels = []
    for name in ['logistic', 'catboost']:
        model = models[name]['model']
        y_test, test_pred_proba = _processTestData(model, data)
        precision, recall, thresholds = precision_recall_curve(
            y_test, test_pred_proba, pos_label=1)
        fscore = (2 * precision * recall) / (precision + recall)
        idx = np.nanargmax(fscore)
        PrecisionRecallDisplay.from_estimator(
            model, data['X_test'], data['y_test'], ax=ax)
        if name == 'catboost':
            ax.axhline(precision[idx], xmax=recall[idx], ls='--', alpha=0.5, c='black')
            ax.axvline(recall[idx], ymax=precision[idx], ls='--', alpha=0.5, c='black')
            ax.scatter(recall[idx], precision[idx], c='black')
        label = f'{name}: F-score = {fscore[idx]:.2f}, Threshold = {thresholds[idx]:.3f}'
        labels.append(label)
    noSkill = data['y_test'].sum() / len(data['y_test'])
    ax.axline((0, noSkill), (1, noSkill), ls='--', color='red')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    ax.legend(labels=labels, loc='upper right')
    return fig, ax


def plotCalibrationCurve(models, data, figsize=None, **kwargs):
    fig, ax = plt.subplots(figsize=figsize)
    names = ['logistic', 'catboost']
    for name in names:
        model = models[name]['model']
        CalibrationDisplay.from_estimator(
            model, data['X_test'], data['y_test'],
            ref_line=False, ax=ax, **kwargs)
    ax.axline((0, 0), slope=1, ls='--', color='red')
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.legend(labels=names, loc='lower right')
    return fig, ax


def predict(model, X):
    """ Generate predictions using trained model """
    try:
        param = 'preprocess__prepare__decisionThreshold'
        threshold = model.get_params()[param]
    except KeyError:
        logger.error('No threshold set - setting to 0.5.')
        threshold = 0.5
    classes = model.classes_
    out = pd.DataFrame(model.predict_proba(X), columns=classes)
    out['class'] = out[classes[1]].apply(
        lambda x: classes[0] if x < threshold else classes[1])
    return out.values


def evaluate(model, data):
    """ Generate classification report using test data """
    predictions = predict(model, data['X_test'])[:,2]
    report = classification_report(
        data['y_test'], predictions, output_dict=True)
    return report



#!/usr/bin/env python3

""" DNAttend """

import sys
import logging
import argparse
from timeit import default_timer as timer
from .main import train_cli, test_cli, retrain_cli, predict_cli, simulate_cli
from ._version import __version__


def parseArgs() -> argparse.Namespace:
    epilog = 'Stephen Richer, NHS England (stephen.richer@nhs.net)'
    baseParser = getBaseParser(__version__)
    parser = argparse.ArgumentParser(
        epilog=epilog, description=__doc__, parents=[baseParser])
    subparser = parser.add_subparsers(
        title='required commands',
        description='',
        dest='command',
        metavar='Commands',
        help='Description:')

    sp1 = subparser.add_parser(
        'train',
        description=train_cli.__doc__,
        help='Train model.',
        parents=[baseParser],
        epilog=parser.epilog)
    sp1.add_argument(
        'config', help='YAML configuration file.')
    sp1.set_defaults(function=train_cli)


    sp2 = subparser.add_parser(
        'test',
        description=test_cli.__doc__,
        help='Test model.',
        parents=[baseParser],
        epilog=parser.epilog)
    sp2.add_argument(
        'config', help='YAML configuration file.')
    sp2.set_defaults(function=test_cli)


    sp3 = subparser.add_parser(
        'retrain',
        description=retrain_cli.__doc__,
        help='Retrain model.',
        parents=[baseParser],
        epilog=parser.epilog)
    sp3.add_argument(
        'config', help='YAML configuration file.')
    sp3.set_defaults(function=retrain_cli)


    sp4 = subparser.add_parser(
        'predict',
        description=predict_cli.__doc__,
        help='Run predictions model.',
        parents=[baseParser],
        epilog=parser.epilog)
    sp4.add_argument(
        'data', help='Data to generate predictions.')
    sp4.add_argument(
        'model', help='Trained model in pickl format.')
    sp4.add_argument(
        '--verify', action='store_true',
        help='Validate output against example hash (default: %(default)s)')
    sp4.add_argument(
        '--encoding', default='utf-8',
        help='Encoding to use for UTF when reading (default: %(default)s)')
    sp4.add_argument(
        '--sep', default=',',
        help='Seperator of input dats (default: %(default)s)')
    sp4.set_defaults(function=predict_cli)


    sp5 = subparser.add_parser(
        'simulate',
        description=simulate_cli.__doc__,
        help='Simulate test data.',
        parents=[baseParser],
        epilog=parser.epilog)
    sp5.add_argument(
        '--config',
        help='Path to write default config file (default: stderr)')
    sp5.add_argument(
        '--size', type=int, default=50_000,
        help='Number of records to simulate (default: %(default)s)')
    sp5.add_argument(
        '--noise', type=float, default=0.2,
        help='Scale factor for random noise (default: %(default)s)')
    sp5.add_argument(
        '--seed', type=int, default=42,
        help='Seed for random number generator (default: %(default)s)')
    sp5.set_defaults(function=simulate_cli)

    args = parser.parse_args()
    if 'function' not in args:
        parser.print_help()
        sys.exit()

    rc = executeCommand(args)
    return rc


def executeCommand(args):
    # Initialise logging
    logFormat = '%(asctime)s - %(levelname)s - %(funcName)s - %(message)s'
    logging.basicConfig(level=args.verbose, format=logFormat)
    logging.captureWarnings(True)
    del args.verbose, args.command
    # Pop main function and excute script
    function = args.__dict__.pop('function')
    start = timer()
    rc = function(**vars(args))
    end = timer()
    logging.info(f'Total execution time: {end - start:.3f} seconds.')
    logging.shutdown()
    return rc


def getBaseParser(version: str) -> argparse.Namespace:
    """ Create base parser of verbose/version. """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        '--version', action='version', version='%(prog)s {}'.format(version))
    parser.add_argument(
        '--verbose', action='store_const', const=logging.INFO,
        default=logging.ERROR, help='verbose logging for debugging')
    return parser



#!/usr/bin/env python3

import yaml
import pprint
import logging
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin


logger = logging.getLogger(__name__)


def setVerbosity(
        level=logging.INFO, handler=logging.StreamHandler(),
        format='%(name)s - %(levelname)s - %(message)s'):
    formatter = logging.Formatter(format)
    handler.setFormatter(formatter)
    pkg_logger = logging.getLogger('dnattend')
    pkg_logger.setLevel(level)
    pkg_logger.addHandler(handler)


class _prepareData(BaseEstimator, TransformerMixin):

    def __init__(
            self,
            catCols: list = None,
            numericCols: list = None,
            boolCols: list = None,
            decisionThreshold: float = 0.5):
        self.catCols = [] if catCols is None else catCols
        self.numericCols = [] if numericCols is None else numericCols
        self.boolCols = [] if boolCols is None else boolCols
        # Store parameter for decision threshold within model object
        self.decisionThreshold = decisionThreshold
        self._setCatColIdx()

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        for col in self.boolCols:
            X[col] = self._mapBoolCol(X[col])
        X[self.catCols] = X[self.catCols].astype(str)
        return X.loc[:, self.validCols]

    def _mapBoolCol(self, col):
        col = col.apply(lambda x: str(x).lower().strip()[0])
        names = ({
            '1': 1, 'y': 1, 't': 1,
            '0': 0, 'n': 0, 'f': 0
        })
        col = col.map(names)
        col.loc[~col.isin([0,1])] = np.nan
        return col

    @property
    def validCols(self):
        return self.catCols + self.numericCols + self.boolCols

    def _setCatColIdx(self):
        """ Get indices of categoric cols """
        self.catColIdx = []
        for col in self.catCols:
            if col in self.validCols:
                self.catColIdx.append(
                    self.validCols.index(col))


class Config():
    """ Custom class to read, validate and
        set defaults of YAML configuration file
        """

    def __init__(self, pathToYaml):
        self.error = False
        # Reserved string for mandatory arguments
        self.mandatory = 'mandatory'
        self.pathToYaml = pathToYaml
        self.config = self._readYAML()
        self._setDefault(self.config, self.default)
        self._postProcessConfig()
        if self.error:
            logging.error(
                f'Expected format:\n\n{pprint.pformat(self.default)}')
            raise ValueError('Invalid configuration.')

    def _setDefault(self, config, default, path=''):
        """ Recursively set default values. """
        for k in default:
            if isinstance(default[k], dict):
                if not isinstance(config[k], dict):
                    logging.error(
                        f'"{config[k]}" should be a dictionary.')
                    self.error = True
                else:
                    self._setDefault(
                        config.setdefault(k, {}), default[k], path=path+k)
            else:
                if (((k not in config) or (config[k] is None))
                        and (default[k] == self.mandatory)):
                    msg = f'{path}: {k}' if path else k
                    logging.error(
                        f'Missing mandatory config "{msg}".')
                    self.error = True
                config.setdefault(k, default[k])

    def _readYAML(self):
        """ Custom validation """
        with open(self.pathToYaml, 'r') as stream:
            return yaml.safe_load(stream)

    @property
    def default(self):
        """ Default values of configuration file. """
        return ({
            'input': self.mandatory,
            'target': self.mandatory,
            'DNAclass': 1,
            'out': '.',
            'finalModel': '',
            'hyperParams': None,
            'catCols': [],
            'boolCols': [],
            'numericCols': [],
            'train_size': 0.7,
            'test_size': 0.15,
            'val_size': 0.15,
            'tuneThresholdBy':     'f1',
            'cvFolds':             5,
            'catboostIterations':  100,
            'hypertuneIterations': 5,
            'evalIterations':      10_000,
            'earlyStoppingRounds': 10,
            'encoding':            'utf-8',
            'seed':                42
        })


    def _postProcessConfig(self):
        """ Additional config modifications """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" CatBoost Classifer for predicting outpatient non-attendance. """

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev


import os
import sys
import glob
from shutil import rmtree
from setuptools import setup, find_namespace_packages, Command


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


def get_info():
    info = {}
    versionPath = glob.glob('src/*/_version.py')[0]
    with open(versionPath) as fp:
        exec(fp.read(), info)
    return info


setup(
    name='dnattend',
    author='Stephen Richer',
    author_email='stephen.richer@nhs.net',
    url='https://github.com/nhsx/dna-risk-predict',
    entry_points={
        'console_scripts': ['dnattend=dnattend.parseArgs:parseArgs']
    },
    python_requires='>=3.8.0',
    install_requires=[
        'shap',
        'scipy',
        'numpy',
        'pyyaml',
        'pandas',
        'joblib',
        'seaborn',
        'catboost',
        'matplotlib',
        'scikit-learn',
        'category-encoders'
    ],
    extras_require={
        'shap': ['shap'],
    },
    license='MIT',
    classifiers=[
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
    ],
    version=get_info()['__version__'],
    description=__doc__,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    packages=find_namespace_packages(where='src'),
    package_dir={'': 'src'},
    zip_safe=False,
    #cmdclass={
    #    'upload': UploadCommand,
    #}
)



#!/usr/bin/env python3

""" Simulate some dummy data for testing """

import yaml
import random
import logging
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def _setProb(x: pd.Series, noise: float = 0.2) -> pd.Series:
    """ Simulate DNA probability with artificial values """
    modifiers = ({
        'day': 0.05,
        'priority': 0.1,
        'firstAppointment': 0.2,
        'consultationMedia': 0.3,
        'speciality': 0.35,
        'site': 0.4,
        'noise': noise
    })
    maxModifier = np.array(list(modifiers.values())).sum() + 0.5
    minModifier = -maxModifier
    if x['day'] in ['Saturday', 'Sunday']:
        modifiers['day'] *= -1
    if x['priority'] == 'Two Week Wait':
        modifiers['priority'] *= -1
    if not x['firstAppointment']:
        modifiers['firstAppointment'] *= -1
    if x['consultationMedia'] == 'In-Person':
        modifiers['consultationMedia'] *= -1
    if x['speciality'] == 'Audiology':
        modifiers['speciality'] *= -1
    if x['site'] == 'Lakeside':
        modifiers['site'] *= -1
    modifiers['noise'] *= x['noise']
    # Get probability score as sum of modifiers
    p = np.array(list(modifiers.values())).sum()
    # Normalise p to [0, 1]
    p = ((p - minModifier) / (maxModifier - minModifier))
    status = np.random.choice(['DNA', 'Attend'], p=[p, 1-p])
    return pd.Series([p, status])


def generateData(size: int = 50_000, seed: int = 42,
                 noise: float = 0.2) -> pd.DataFrame:
    np.random.seed(seed)
    daysOfWeek = ([
        'Wednesday', 'Tuesday', 'Monday', 'Sunday',
        'Saturday', 'Friday', 'Thursday'
    ])
    logger.info(f'Simulating random dataset with {size} records.')
    data = pd.DataFrame({
        'day': np.random.choice(daysOfWeek, size),
        'priority': np.random.choice(['Urgent', 'Two Week Wait'], size),
        'age': np.random.choice(range(100), size),
        'speciality': np.random.choice(['Ophthalmology', 'Audiology'], size),
        'firstAppointment': np.random.choice([True, False], size),
        'consultationMedia': np.random.choice(['Remote', 'In-Person'], size),
        'site': np.random.choice(['Fairview', 'Lakeside'], size),
        'noise': np.random.uniform(-1, 0, size) # Modifier for attendance weight
    })
    logger.info(f'Setting target status probabilistically.')
    data[['probability', 'status']] = data.apply(
        _setProb, args=(noise,), axis=1)
    data['status'] = (data['status'] == 'DNA').astype(int)
    return data


def writeConfig(config: str = None):
    config_settings = (''
        'input: DNAttend-example.csv    # Path to input data (Mandatory).\n'
        'target: status                 # Column name of target (Mandatory).\n'
        'DNAclass: 1                    # Value of target corresponding to DNA.\n'
        'out: .                         # Output directory to save results.\n'
        'finalModel: catboost           # Method to train final model (catboost or logistic).\n'
        'catCols:                       # Column names of categorical features.\n'
        '    - day\n'
        '    - priority\n'
        '    - speciality\n'
        '    - consultationMedia\n'
        '    - site\n'
        'boolCols:                      # Column names of boolean features.\n'
        '    - firstAppointment\n'
        'numericCols:                   # Column names of numeric features.\n'
        '    - age\n'
        'train_size: 0.7                # Proportion of data for training.\n'
        'test_size: 0.15                # Proportion of data for testing.\n'
        'val_size: 0.15                 # Proportion of data for validation.\n'
        'tuneThresholdBy: f1            # Metric to tune decision threshold (f1 or roc).\n'
        'cvFolds: 5                     # Hyper-tuning cross-validations.\n'
        'catboostIterations: 100        # Hyper-tuning CatBoost iterations.\n'
        'hypertuneIterations: 5         # Hyper-tuning parameter samples.\n'
        'evalIterations: 10_000         # Upper-limit over-fit iterations.\n'
        'earlyStoppingRounds: 10        # Over-fit detection early stopping rounds.\n'
        'encoding: latin-1              # Encoding to use for reading files.\n'
        'seed: 42                       # Seed to ensure workflow reproducibility.\n'
    )
    if config is None:
        sys.stderr.write(config_settings)
    else:
        with open(config, 'w') as fh:
            fh.write(config_settings)
